﻿/*
 * Created by SharpDevelop.
 * User: Malte Schulze
 * Date: 31.05.2017
 * Time: 15:57
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using RIMMSqol.genericSettings;
using RIMMSqol.pathfinding;
using RIMMSqol.renderers;
using RIMMSqol.extensions;
using RimWorld;
using Verse;
using UnityEngine;
using Verse.AI;
using System.IO;
using System.Xml;
using HarmonyLib;

namespace RIMMSqol
{
	/// <summary>
	/// Description of QOLMod.
	/// </summary>
	public class QOLMod : Mod
	{
		//utilities that are missing in the native code
		protected static float[] lineHeights;
		public static float LineHeight(GameFont font) {
			return lineHeights[(int)font];
		}
		public static GUIStyle FontStyle(GameFont font, TextAnchor anchor = TextAnchor.UpperLeft, bool wordWrap = true) {
			GUIStyle gUIStyle = Text.fontStyles[(int)font];
			gUIStyle.alignment = anchor;
			gUIStyle.wordWrap = wordWrap;
			return gUIStyle;
		}
		protected static GUIContent tmpTextGUIContent;
		public static Vector2 CalcSize(string text, GameFont font, TextAnchor anchor = TextAnchor.UpperLeft, bool wordWrap = true)
		{
			if ( text == null ) return Vector2.zero;
			tmpTextGUIContent.text = text;
			return FontStyle(font, anchor, wordWrap).CalcSize(tmpTextGUIContent);
		}
		public static float CalcHeight(string text, float width, GameFont font, TextAnchor anchor = TextAnchor.UpperLeft)
		{
			if ( text == null ) return 0f;
			tmpTextGUIContent.text = text;
			return FontStyle(font, anchor, true).CalcHeight(tmpTextGUIContent, width);
		}
		protected static MethodInfo _GUIStyleGetNumCharactersThatFitWithinWidth = typeof(GUIStyle).GetMethod("GetNumCharactersThatFitWithinWidth",BindingFlags.NonPublic|BindingFlags.Instance);
		public static string CapString(string text, float width, string suffix, out bool wasCapped, GameFont font, TextAnchor anchor = TextAnchor.UpperLeft) {
			if ( text == null ) {
				wasCapped = false;
				return null;
			}
			if ( text.Contains('\n') ) {
				wasCapped = true;
				bool wasCappedSecondary;
				return CapString(text.Substring(0,text.FirstIndexOf(c=>c=='\n')), width, suffix, out wasCappedSecondary, font, anchor);
			}
			float widthForText = CalcSize(text,font,anchor,false).x;
			if ( width >= widthForText ) {
				wasCapped = false;
				return text;
			}
			float widthForSuffix;
			if ( suffix.NullOrEmpty() ) widthForSuffix = 0;
			else widthForSuffix = CalcSize(suffix,font,anchor,false).x;
			int charsToKeep = (int)_GUIStyleGetNumCharactersThatFitWithinWidth.Invoke(FontStyle(font, anchor, false),new object[]{text,Mathf.Max(0,width-widthForSuffix)});
			wasCapped = text.Length != charsToKeep;
			if ( widthForSuffix > 0 ) return text.Substring(0,charsToKeep)+suffix;
			else return text.Substring(0,charsToKeep);
		}
		private static float _verticalScrollbarWidth = -1;
		public static float VerticalScrollbarWidth() {
			if ( _verticalScrollbarWidth > 0 ) return _verticalScrollbarWidth;
			try {
				GUIStyle vScrollStyle = GUI.skin.verticalScrollbar;
				//UnityEngine ignores margin right for scrollbar, this is the exact width taken up by a vertical scrollbar.
				_verticalScrollbarWidth = vScrollStyle.fixedWidth + (float)vScrollStyle.margin.left;
				return _verticalScrollbarWidth;
			} catch {
				//In case this is called outside of a GUI context the skin can not be retrieved
				return 16f;
			}
		}
		private static float _horizontalScrollbarHeight = -1;
		public static float HorizontalScrollbarHeight() {
			if ( _horizontalScrollbarHeight > 0 ) return _horizontalScrollbarHeight;
			try {
				GUIStyle hScrollStyle = GUI.skin.horizontalScrollbar;
				_horizontalScrollbarHeight = hScrollStyle.fixedHeight + (float)hScrollStyle.margin.top;
				return _horizontalScrollbarHeight;
			} catch {
				//In case this is called outside of a GUI context the skin can not be retrieved
				return 16f;
			}
		}
		private static Dictionary<Color,Texture2D> cachedSolidColorTextures = new Dictionary<Color, Texture2D>();
		public static Texture2D getSolidColorTexture(Color color) {
			Texture2D tex;
			if ( !cachedSolidColorTextures.TryGetValue(color, out tex) ) {
				tex = SolidColorMaterials.NewSolidColorTexture(color);
				cachedSolidColorTextures.Add(color,tex);
			}
			return tex;
		}
		private QOLModSettings settings;
		private static List<Action<Mod>> applySettingsListeners = new List<Action<Mod>>();
		private static List<Action<Mod>> preApplySettingsListeners = new List<Action<Mod>>();
		private static QOLMod singleton;
		private Flow flow;
		private static Pawn dummyPawn;
		private bool skipFlowInitOnce = false;
		public static Pawn getDummyPawn() {
			//minimalistic pawn that is used to generate tooltips
			if ( dummyPawn == null ) {
				dummyPawn = new Pawn();
				dummyPawn.def = DefDatabase<ThingDef>.GetNamed("Human",true);
				dummyPawn.kindDef = DefDatabase<PawnKindDef>.GetNamed("Colonist");
				dummyPawn.ageTracker = new Pawn_AgeTracker(dummyPawn);
				dummyPawn.records = new Pawn_RecordsTracker(dummyPawn);
				dummyPawn.gender = Gender.Male;
				dummyPawn.Name = new NameSingle("描きました");
			}
			return dummyPawn;
		}
		
		public QOLMod(ModContentPack content) : base(content)
		{
            //creating the merge file directories if necessary
            MergeFile.GetMergeFileDirectory(this);
            //settings = GetSettings<QOLModSettings>();
            //In case the configuration file was not loaded but newly created we want to init the settings here.
            //SettingsInit.init();
            //Registering the xml processing early
            var harmony = new Harmony("RIMMSqol");
            List<HarmonyMethod> harmonyMethods = HarmonyMethodExtensions.GetFromType(typeof(LoadedModManager_ParseAndProcessXML));
            new PatchClassProcessor(harmony, typeof(LoadedModManager_ParseAndProcessXML)).Patch();

            flow = null;
			singleton = this;
			System.Reflection.FieldInfo fi = typeof(Text).GetField("lineHeights", System.Reflection.BindingFlags.NonPublic | System.Reflection.BindingFlags.Static);
			lineHeights = (float[])fi.GetValue(null);
			tmpTextGUIContent = new GUIContent();
        }

		public override string SettingsCategory()
		{
			return "QoL";
		}
		
		public override void WriteSettings()
		{
			if ( flow != null ) {
				flow.destroy();
				flow = null;
				//The way the events are processed a final rendering step takes place before the window is actually removed. We don't wan't to recreate the flow on that step.
				//If this is ever fixed in the base game we would only skip one frame the next time the window is opened.
				skipFlowInitOnce = true;
			}
			try {
				ApplySettings_();
            } catch (Exception ex) {
				Log.Error("Broad error while applying settings. "+ex.Message+"\n"+ex.StackTrace);
			}
            //delaying the writing of the settings so that forbidden patches and defcopies can copy the user input into their variables through merging
            base.WriteSettings();
        }
		
		public static void addApplySettingsListener(Action<Mod> callback) {
			applySettingsListeners.Add(callback);
		}
		
		public static void addPreApplySettingsListener(Action<Mod> callback) {
			preApplySettingsListeners.Add(callback);
		}
		
		public static bool stopSkillDecay() {
			return singleton.settings.stopSkillDecay;
		}

        public static bool stopSkillDown()
        {
            return singleton.settings.stopSkillDown;
        }

        public static bool stopRoofCollapse() {
			return singleton.settings.stopRoofCollapse;
		}
		
		public static int skillDecayLevelOffset() {
			return singleton.settings.skillDecayLevelOffset;
		}
		
		public static bool stopTamenessDecay() {
			return singleton.settings.stopTamenessDecay;
		}

        public static bool displayFoundCityGizmo() {
			return singleton.settings.displayFoundCityGizmo;
		}
		
		public static bool preventAnimalFamilies(Pawn p1, Pawn p2) {
            QOLModSettings settings = getSettings();
            if (!settings.preventAnimalFamilies) return false;
            if (settings.allowNamedAnimalFamilies)
            {
                if (p1?.Name != null && !p1.Name.Numerical) return false;
                if (p2?.Name != null && !p2.Name.Numerical) return false;
            }

            return true;
		}
		
		public static bool useFixedNumericTextfields() {
			return singleton.settings.useFixedNumericTextfields;
		}

        public void InitializeSettings()
        {
            typeof(Mod).GetField("modSettings", BindingFlags.Instance | BindingFlags.NonPublic | BindingFlags.Public).SetValue(this, null);
            settings = GetSettings<QOLModSettings>();
        }
		
		public static void ApplySettings(bool reload = false) {
            singleton.ApplySettings_(reload);
		}
		
		public static QOLModSettings getSettings() {
			return singleton.settings;
		}
		
		public static QOLMod getModHandle() {
			return singleton;
		}
		
		public static float getRemnantOrderPriceFactor() {
			return singleton.settings.remnantOrderPriceFactor;
		}

        public static float getTooltipMaxWidth()
        {
            return singleton.settings.tooltipMaxWidth;
        }

        protected static Dictionary<int,ReflectionMethodInvoker> pfAlgorithms = new Dictionary<int, ReflectionMethodInvoker>();
		
		public static void setPFAlgorithms(Dictionary<int,ReflectionMethodInvoker> pfAlgorithms) {
			if ( pfAlgorithms != null ) QOLMod.pfAlgorithms = pfAlgorithms;
		}
		
		public static void reloadPFAlgorithms() {
			foreach ( ReflectionMethodInvoker rmi in pfAlgorithms.Values ) {
				try {
					MethodInfo miReload = rmi.instance.GetType().GetMethod("Reload");
					if ( miReload != null ) {
						miReload.Invoke(rmi.instance, null);
					}
				} catch {
					Log.Warning("Failed to determine Reload method on pathfinding algorithm \""+rmi.instance.GetType()+"\".");
				}
			}
		}
		
		public static bool hasPFAlgorithm(int algId) {
			return pfAlgorithms.ContainsKey(algId);
		}
		
		public static PawnPath doPFAlgorithm(int algId, object[] parameters) {
			ReflectionMethodInvoker invoker;
			if ( pfAlgorithms.TryGetValue(algId,out invoker) ) {
				return (PawnPath)invoker.invoke(parameters);
			}
			return null;
		}
		
		public static PawnPath doPFAlgorithm(int algId, Map map, IntVec3 start, LocalTargetInfo dest, TraverseParms traverseParms, PathEndMode peMode, ByteGrid avoidGrid, Area allowedArea,
		                        int costsCardinal, int costsDiagonal) {
			ReflectionMethodInvoker invoker;
			if ( pfAlgorithms.TryGetValue(algId,out invoker) ) {
				return invoker.findPath(map,start,dest,traverseParms,peMode,avoidGrid,allowedArea,costsCardinal,costsDiagonal);
			}
			return null;
		}
		
		public static float getTraitCost(Trait t, bool removal = false) {
			return getTraitCost(t.def.defName,t.CurrentData.degree,removal);
		}
		
		public static float getTraitCost(string traitDefName, int degree, bool removal = false) {
            if (singleton.settings.useSyrchalistTraitvalue && ModsConfig.IsActive("syrchalis.traitvalue"))
            {
                Type tTraitValueExtension = DefSettingsField.GetTypeFromAnyAssemblyVersion("SyrTraitValue.TraitValueExtension, TraitValue, Version=1.0.0.0, Culture=neutral, PublicKeyToken=null");
                if (tTraitValueExtension != null)
                {
                    try
                    {
                        TraitDef trait = DefDatabase<TraitDef>.GetNamed(traitDefName);
                        foreach (DefModExtension me in trait.modExtensions)
                        {
                            if (me != null && tTraitValueExtension.IsAssignableFrom(me.GetType()))
                            {
                                IList lst = (IList)tTraitValueExtension.GetField("traitValues").GetValue(me);
                                if (lst != null)
                                {
                                    Type tDegreeValue = DefSettingsField.GetTypeFromAnyAssemblyVersion("SyrTraitValue.DegreeValue, TraitValue, Version=1.0.0.0, Culture=neutral, PublicKeyToken=null");
                                    FieldInfo fDegree = tDegreeValue.GetField("degree");
                                    foreach (object o in lst)
                                    {
                                        if ((int)fDegree.GetValue(o) == degree)
                                        {
                                            int value = (int)tDegreeValue.GetField("value").GetValue(o);
                                            return value / 15.0f;
                                        }
                                    }
                                }
                            }
                        }
                    } catch (Exception ex)
                    {
                        Log.Error("Unable to check [Syr] Trait Value mod for trait cost!\n"+ex.StackTrace);
                    }
                }
            }

            QOLTraitCostDef defaultCost = DefDatabase<QOLTraitCostDef>.GetNamed("QOLTraitCost"+traitDefName+degree,false);
			if ( defaultCost != null ) {
				if (!removal || Math.Abs(defaultCost.costRemoval) < float.Epsilon ) return defaultCost.cost;
				else return defaultCost.costRemoval;
			}
			return 2;
		}
		
		public static float getCuMRecordFactor(RecordDef def) {
			QOLRecordFactorDef recordFactorDef = DefDatabase<QOLRecordFactorDef>.GetNamed("QOLRecordFactor"+def.defName,false);
			if ( recordFactorDef != null ) return recordFactorDef.factor;
			return 0;
		}
		
		public static float getCuMRecordPoints(Pawn p, RecordDef def) {
			return p.records.GetValue(def) * getCuMRecordFactor(def);
		}
		
		public static float getBaseCost() {
			return singleton.settings.cumBaseCost;
		}

        public static float getCuMCostIdeo()
        {
            return singleton.settings.cumCostIdeo;
        }

        public static float getCumPointPawnToPoolConversionFactor() {
			return singleton.settings.cumPointPawnToPoolConversionFactor;
		}
		
		public static float getCumPointRemnantsToPoolConversionFactor() {
			return singleton.settings.cumPointRemnantsToPoolConversionFactor;
		}
 
		private void ApplySettings_(bool reload = false) {
			//in case there is no mod configuration to load we initialize the settings here, so that the mod menu can be used without errors.
			SettingsInit.init();
			if (settings == null) return;
			
			foreach(Action<Mod> listener in preApplySettingsListeners ) {
				listener(this);
			}
			
			SettingsStorage.ApplySettings(reload);
			
			foreach(Action<Mod> listener in applySettingsListeners ) {
				listener(this);
			}
		}
		
		protected static SettingsProperties navigateToProps; 
		protected static object navigateToEntry;
		
		public static void NavigateToEditing(SettingsProperties props, object entry) {
			navigateToProps = props;
			navigateToEntry = entry;
		}
		
		public override void DoSettingsWindowContents(Rect inRect)
		{
			if ( skipFlowInitOnce ) {
				skipFlowInitOnce = false;
				return;
			}
			
			List<ISettingsInstance> lstForNavigation = null;
			if ( flow == null ) {
				Dictionary<SettingsProperties,List<ISettingsInstance>> lstsSummary = new Dictionary<SettingsProperties, List<ISettingsInstance>>();
				List<Pair<string,Action<Flow>>> buttonInstructions = new List<Pair<string, Action<Flow>>>();
				Dictionary<string,PageRenderer> pages = new Dictionary<string, PageRenderer>();
				pages.Add("menu", new PageRenderer((f=>"Menu"), null, null)
				          .AddChild(new FilteredGridListRenderer<Pair<string,Action<Flow>>>(buttonInstructions,-1,-1)
                                     .AddChild(new ButtonTextRenderer(f=>((Pair<string,Action<Flow>>)((IterationItem)f.pageScope["curItem"]).curItem).First,f=>((Pair<string,Action<Flow>>)((IterationItem)f.pageScope["curItem"]).curItem).Second(f), GameFont.Small, 200f))));
				
				foreach ( SettingsProperties props in SettingsStorage.getAllSettingsProperties() ) {
					if ( props.visualization.displayInMenu ) {
						if ( props.visualization.displaySelection ) {
							pages.Add("edit"+props.id,editPage(props,"select"+props.id,"edit"+props.id,pages));
							List<ISettingsInstance> lst = SettingsStorage.generateListForSelection(props);
							lstsSummary.Add(props,lst);
							if ( navigateToProps == props ) lstForNavigation = lst;
							pages.Add("select"+props.id,selectPage(((ISettingsVisualizationSelection)props.visualization).selectPageTitle, "menu", lst, "edit"+props.id));
							buttonInstructions.Add(new Pair<string, Action<Flow>>(((ISettingsVisualizationMenu)props.visualization).mainMenuButtonLabel, f => f.navigate("select"+props.id)));
						} else {
							//editing a singleton object means we pass over selection but set the selected object to the singleton instance
							pages.Add("edit"+props.id,editPage(props,"menu","edit"+props.id,pages));
							ISettingsInstance inst = SettingsStorage.generateSingletonObject(props);
							buttonInstructions.Add(new Pair<string, Action<Flow>>(((ISettingsVisualizationMenu)props.visualization).mainMenuButtonLabel, 
                                      f => {f.flowScope["edit"] = inst; f.navigate("edit"+props.id); }));
						}
					}
				}
				
				List<Pair<ISettingsInstance,Action<Flow>>> summaryInstructions = new List<Pair<ISettingsInstance, Action<Flow>>>();
				foreach (SettingsProperties prop in lstsSummary.Keys ) {
					List<ISettingsInstance> insts = lstsSummary[prop];
					string pageId = "edit"+prop.id;
					foreach (ISettingsInstance inst in insts ) {
						summaryInstructions.Add(new Pair<ISettingsInstance, Action<Flow>>(inst, f=>{
						                                                       	flow.flowScope["edit"] = inst;
																				flow.navigate(pageId);
						                                                       }));
					}
				}
				lstsSummary.Clear();
				string summaryTitle = "Summary";
				summaryTitle.TryTranslate(out summaryTitle);
				pages.Add("summary", new PageRenderer(f=>summaryTitle,null,()=>"menu")
				          .AddChild(new FilteredGridListRenderer<Pair<ISettingsInstance,Action<Flow>>>(summaryInstructions,-1,-1,(filterExpr, item) => item.First.getActive() && filterMatchesSettingsInstance(filterExpr, item.First))
                                     .AddChild(new ButtonTextRenderer(f=>((Pair<ISettingsInstance,Action<Flow>>)((IterationItem)f.pageScope["curItem"]).curItem).First.getLabel(),f=>((Pair<ISettingsInstance,Action<Flow>>)((IterationItem)f.pageScope["curItem"]).curItem).Second(f), GameFont.Small, 200f))
                                     .AddChild(new ButtonRenderer(Widgets.CheckboxOffTex, 
				                                       (f => {((Pair<ISettingsInstance,Action<Flow>>)((IterationItem)f.pageScope["curItem"]).curItem).First.setActive(false); ((Pair<ISettingsInstance,Action<Flow>>)((IterationItem)f.pageScope["curItem"]).curItem).First.reset(); }), 24f, 24f)
				                                       .hidden((f => !((Pair<ISettingsInstance,Action<Flow>>)((IterationItem)f.pageScope["curItem"]).curItem).First.getActive())))
				                   ));
				buttonInstructions.Add(new Pair<string, Action<Flow>>("Summary", f => f.navigate("summary")));

                buttonInstructions.Add(new Pair<string, Action<Flow>>("Import/Export", f =>
                {
                    //rimworlds long event handling is bugged in that it scrambles the window order after displaying an immediate window on the super layer.
                    //we close all other windows to prevent issues.
                    foreach ( Window wnd in Find.WindowStack.Windows.ToList() )
                    {
                        wnd.Close(false);
                    }
                    Find.WindowStack.Add(new Dialog_ImportExportSettings());
                }));

                flow = new Flow(pages,"menu");
			}
			
			if ( navigateToProps != null ) {
				if ( navigateToProps.visualization.displaySelection ) {
					if ( lstForNavigation == null ) Log.Error("Failed to find corresponding list for preset navigation into props "+navigateToProps.id);
					else {
                        if (navigateToEntry != null)
                        {
                            string key = null;
                            try
                            {
                                key = ((ISettingsVisualizationSelection)navigateToProps.visualization).getKeyFromBaseObject(navigateToEntry);
                            }
                            catch
                            {
                                Log.Error("Failed to get key for preset navigation. Props: " + navigateToProps.id + " Baseobject: " + navigateToEntry);
                            }
                            if (key != null)
                            {
                                ISettingsInstance inst = lstForNavigation.Find(si => si.getKey().Equals(key));
                                if (inst != null)
                                {
                                    flow.flowScope["edit"] = inst;
                                    flow.navigate("edit" + navigateToProps.id);
                                }
                                else Log.Error("Failed to find settings for key: " + key + " in props: " + navigateToProps.id);
                            }
                        }
                        else
                        {
                            flow.navigate("select" + navigateToProps.id);
                        }
					}
				} else {
					flow.flowScope["edit"] = SettingsStorage.generateSingletonObject(navigateToProps); 
					flow.navigate("edit"+navigateToProps.id);
				}
				navigateToProps = null;
				navigateToEntry = null;
			}
			flow.DoFlowContents(inRect);
		}

        protected bool filterMatchesSettingsInstance(string filterExpr, ISettingsInstance item)
        {
            if (filterExpr.NullOrEmpty()) return true;
            if (filterExpr.Contains("#"))
            {
                Def def = ((ISettingsVisualizationSelection)SettingsStorage.getSettingsProperties(item.getId()).visualization).getBaseObjectFromKey(item.getKey()) as Def;
                if (def != null && def.modContentPack != null && def.modContentPack.Name != null)
                {
                    string[] parts = filterExpr.Split(new char[] { '#' }, 2);
                    if (parts.Length == 2)
                    {
                        return (parts[0].Length == 0 || item.getLabel().IndexOf(parts[0], StringComparison.OrdinalIgnoreCase) >= 0) &&
                            (parts[1].Length == 0 || def.modContentPack.Name.IndexOf(parts[1], StringComparison.OrdinalIgnoreCase) >= 0);
                    }
                }
            }
            return item.getLabel().IndexOf(filterExpr, StringComparison.OrdinalIgnoreCase) >= 0;
        }

        protected PageRenderer selectPage(string pageTitle, string backPage, List<ISettingsInstance> items, string editPage) {
            return new PageRenderer(f => pageTitle, null, () => backPage).AddChild(new FilteredGridListRenderer<ISettingsInstance>(items, -1f, -1f, filterMatchesSettingsInstance) { SkipOffscreenChildren = true }.AddChild(new ButtonTextRenderer(f => ((ISettingsInstance)((IterationItem)f.pageScope["curItem"]).curItem).getLabel(), (f => {f.flowScope["edit"] = ((IterationItem)f.pageScope["curItem"]).curItem; f.navigate(editPage); } ), GameFont.Small, 200f))
				.AddChild(new ButtonRenderer(Widgets.CheckboxOffTex, (f => {((ISettingsInstance)((IterationItem)f.pageScope["curItem"]).curItem).setActive(false); ((ISettingsInstance)((IterationItem)f.pageScope["curItem"]).curItem).reset(); }), 24f, 24f)
				        .hidden((f => !((ISettingsInstance)((IterationItem)f.pageScope["curItem"]).curItem).getActive()))));
		}
		
		protected PageRenderer editPage(SettingsProperties props, string backPage, string currentPage, Dictionary<string, PageRenderer> pages, Action<Flow,PageRenderer> onNavigationHandler = null) {
            RowLayoutRenderer rows = new RowLayoutRenderer();
			List<ISettingsFieldProperties> fis = props.fields;
			foreach ( ISettingsFieldProperties p in fis ) {
                int startChildCount = rows.childs.Count;
				
				if ( !p.isList && p.isSelectable ) {
                    if ( p is SettingsFieldPropertiesSelectableIcon )
                    {
                        SettingsFieldPropertiesSelectableIcon pSelect = (SettingsFieldPropertiesSelectableIcon)p;
                        rows.AddChild(new ColumnLayoutRenderer()
                                .AddChild(new LabelRenderer(pSelect.label, 200f, GameFont.Small, TextAnchor.UpperLeft).attachTooltip(p))
                                .AddChild(new ButtonRenderer(f => pSelect.iconProducer(((ISettingsInstance)f.flowScope["edit"]).get<object>(p.id)),
                                                                f => Find.WindowStack.Add(new Dialog_Select<object>(pSelect.onSelectItem(f), pSelect.selectableItems(f), pSelect.labelProducer, pSelect.iconProducer)),
                                                                   32f, 32f)
                                { DrawBackground = true })
                                .AddChild(new ColumnLayoutRenderer().setPaddingBetweenColumns(0f)
                                    .AddChild(new HideCollapseRenderer(f => ((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id), new ButtonRenderer(Materials.radioButtonOff, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsConfigured(p.id), 24f, 24f)))
                                    .AddChild(new HideCollapseRenderer(f => !((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id) || !((ISettingsInstance)f.flowScope["edit"]).hasFieldConfiguredStatus(), new ButtonRenderer(Materials.radioButtonOn, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsNotConfigured(p.id), 24f, 24f)))
                                ));
                    }
                    else
                    {
                        ISettingsFieldPropertiesSelectSingle pSelect = (ISettingsFieldPropertiesSelectSingle)p;
                        rows.AddChild(new ColumnLayoutRenderer()
                                      .AddChild(new LabelRenderer(pSelect.label, 200f).attachTooltip(p))
                                .AddChild(new ButtonTextRenderer(f => pSelect.labelProducer(((ISettingsInstance)f.flowScope["edit"]).get<object>(p.id)),
                                                                f => Find.WindowStack.Add(new Dialog_Select<object>(pSelect.onSelectItem(f), pSelect.selectableItems(f), pSelect.labelProducer, pSelect.iconProducer)),
                                                                   GameFont.Small, 200f))
                                .AddChild(new ColumnLayoutRenderer().setPaddingBetweenColumns(0f)
                                    .AddChild(new HideCollapseRenderer(f => ((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id), new ButtonRenderer(Materials.radioButtonOff, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsConfigured(p.id), 24f, 24f)))
                                    .AddChild(new HideCollapseRenderer(f => !((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id) || !((ISettingsInstance)f.flowScope["edit"]).hasFieldConfiguredStatus(), new ButtonRenderer(Materials.radioButtonOn, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsNotConfigured(p.id), 24f, 24f)))
                                ));
                    }
				} else if ( !p.isList && p.isSettings ) {
					SettingsFieldPropertiesSettings pSetting = (SettingsFieldPropertiesSettings)p;
					SettingsProperties props2 = SettingsStorage.getSettingsProperties(pSetting.idEnclosedSettings);
					string deepEditPageId = currentPage + "edit"+pSetting.id+props2.id;
					rows.AddChild(new ColumnLayoutRenderer()
                                .AddChild(new LabelRenderer(p.label, 200f).attachTooltip(p))
					            .AddChild(new ButtonTextRenderer(f => {
					                                             	SettingsInstance si = ((ISettingsInstance)f.flowScope["edit"]).get<SettingsInstance>(p.id);
					                                             	if ( si != null ) return si.getLabel();
					                                             	else return "NONE";
					                                             },
					                                               f=>{
					                                             	SettingsInstance si = ((ISettingsInstance)f.flowScope["edit"]).get<SettingsInstance>(p.id);
					                                             	if ( si == null ) {
                                                                        if (pSetting.createNewEntry != null) si = (SettingsInstance)pSetting.createNewEntry();
					                                             		else si = (SettingsInstance)p.defaultValue;
					                                             		if ( si != null ) 
					                                             			((ISettingsInstance)f.flowScope["edit"]).set<SettingsInstance>(p.id, si);
					                                             	}
					                                             	if ( si != null ) {
					                                               		f.addPostRenderCallback(fl=>fl.flowScope["edit"] = si);
					                                               		f.navigate(deepEditPageId);
					                                             	} else {
					                                             		Log.Warning("Can not edit settings that are null and have no non null default value.");
					                                             	}
					                                               }, GameFont.Small, 200f - (pSetting.allowRemoval ? 24f : 0f)).attachTooltip(p))
                                                                   .AddChild(pSetting.allowRemoval ? new ButtonTextRenderer(f=> {
                                                                       ISettingsInstance setting = (ISettingsInstance)f.flowScope["edit"];
                                                                       if (((ISettingsInstance)f.flowScope["edit"]).get<SettingsInstance>(p.id) == null)
                                                                       {
                                                                           if (pSetting.createNewEntry != null) return "+";
                                                                           else return "";
                                                                       }
                                                                       else return "-";
                                                                       }, f =>
                                                                   {
                                                                       ISettingsInstance setting = (ISettingsInstance)f.flowScope["edit"];
                                                                       SettingsInstance si = ((ISettingsInstance)f.flowScope["edit"]).get<SettingsInstance>(p.id);
                                                                       if ( si != null )
                                                                       {
                                                                           ((ISettingsInstance)f.flowScope["edit"]).set<SettingsInstance>(p.id, null);
                                                                       }
                                                                       else if(pSetting.createNewEntry != null)
                                                                       {
                                                                           ((ISettingsInstance)f.flowScope["edit"]).set<SettingsInstance>(p.id, (SettingsInstance)pSetting.createNewEntry());
                                                                       }
                                                                   }, GameFont.Small, 20f) : null)
					           .AddChild(new ColumnLayoutRenderer().setPaddingBetweenColumns(0f)
				            	.AddChild(new HideCollapseRenderer(f=>((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id),new ButtonRenderer(Materials.radioButtonOff, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsConfigured(p.id), 24f, 24f)))
			            		.AddChild(new HideCollapseRenderer(f=>!((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id)||!((ISettingsInstance)f.flowScope["edit"]).hasFieldConfiguredStatus(),new ButtonRenderer(Materials.radioButtonOn, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsNotConfigured(p.id), 24f, 24f)))
			            	)
			            );
					
					pages.Add(deepEditPageId,editPage(props2,currentPage,deepEditPageId,pages,(f,nextPage)=>{
					                                  	if ( nextPage.onBack() != deepEditPageId ) {
						                                  	f.flowScope["edit"] = ((ISettingsInstance)f.flowScope["edit"]).getParentObject();
					                                  	}
					                                  }));
				} else if ( p.type == typeof(bool) ) {
					rows.AddChild(new ColumnLayoutRenderer()
			            	.AddChild(new LabelRenderer(p.label, 200f).attachTooltip(p))
			            	.AddChild(new CheckboxRenderer(f => ((ISettingsInstance)f.flowScope["edit"]).get<bool>(p.id), (f,v) => ((ISettingsInstance)f.flowScope["edit"]).set(p.id, v)))
			            	.AddChild(new EmptyRenderer(200f-28f))
			            	.AddChild(new ColumnLayoutRenderer().setPaddingBetweenColumns(0f)
				            	.AddChild(new HideCollapseRenderer(f=>((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id),new ButtonRenderer(Materials.radioButtonOff, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsConfigured(p.id), 24f, 24f)))
			            		.AddChild(new HideCollapseRenderer(f=>!((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id)||!((ISettingsInstance)f.flowScope["edit"]).hasFieldConfiguredStatus(),new ButtonRenderer(Materials.radioButtonOn, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsNotConfigured(p.id), 24f, 24f)))
			            	)
			            );
				} else if ( p.type == typeof(int) ) {
					rows.AddChild(new ColumnLayoutRenderer()
				            .AddChild(new LabelRenderer(p.label, 200f).attachTooltip(p))
				            .AddChild(new EditNumericRenderer<int>(f=>((ISettingsInstance)f.flowScope["edit"]).get<int>(p.id),(f,v) => ((ISettingsInstance)f.flowScope["edit"]).set(p.id, v),
																200f,((ISettingsFieldPropertiesPrimitiveInt)p).minValue,((ISettingsFieldPropertiesPrimitiveInt)p).maxValue))
				            .AddChild(new ColumnLayoutRenderer().setPaddingBetweenColumns(0f)
				            	.AddChild(new HideCollapseRenderer(f=>((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id),new ButtonRenderer(Materials.radioButtonOff, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsConfigured(p.id), 24f, 24f)))
			            		.AddChild(new HideCollapseRenderer(f=>!((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id)||!((ISettingsInstance)f.flowScope["edit"]).hasFieldConfiguredStatus(),new ButtonRenderer(Materials.radioButtonOn, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsNotConfigured(p.id), 24f, 24f)))
			            	)
				           );
				} else if ( p.type == typeof(float) ) {
					rows.AddChild(new ColumnLayoutRenderer()
				            .AddChild(new LabelRenderer(p.label, 200f).attachTooltip(p))
				            .AddChild(new EditNumericRenderer<float>(f=>((ISettingsInstance)f.flowScope["edit"]).get<float>(p.id),(f,v) => ((ISettingsInstance)f.flowScope["edit"]).set(p.id, v),
			                    200f,((ISettingsFieldPropertiesPrimitiveFloat)p).minValue,((ISettingsFieldPropertiesPrimitiveFloat)p).maxValue))
				            .AddChild(new ColumnLayoutRenderer().setPaddingBetweenColumns(0f)
				            	.AddChild(new HideCollapseRenderer(f=>((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id),new ButtonRenderer(Materials.radioButtonOff, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsConfigured(p.id), 24f, 24f)))
			            		.AddChild(new HideCollapseRenderer(f=>!((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id)||!((ISettingsInstance)f.flowScope["edit"]).hasFieldConfiguredStatus(),new ButtonRenderer(Materials.radioButtonOn, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsNotConfigured(p.id), 24f, 24f)))
			            	)
				           );
				} else if ( p.type == typeof(string) ) {
                    if (p is ISettingsFieldPropertiesPrimitiveText)
                    {
                        rows.AddChild(new ColumnLayoutRenderer()
                            .AddChild(new LabelRenderer(p.label, 400f, GameFont.Small, TextAnchor.MiddleCenter).attachTooltip(p))
                            .AddChild(new EmptyRenderer(0f))
                            .AddChild(new ColumnLayoutRenderer().setPaddingBetweenColumns(0f)
                                .AddChild(new HideCollapseRenderer(f => ((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id), new ButtonRenderer(Materials.radioButtonOff, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsConfigured(p.id), 24f, 24f)))
                                .AddChild(new HideCollapseRenderer(f => !((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id) || !((ISettingsInstance)f.flowScope["edit"]).hasFieldConfiguredStatus(), new ButtonRenderer(Materials.radioButtonOn, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsNotConfigured(p.id), 24f, 24f)))
                            )
                           );

                        rows.AddChild(new TextRenderer(f => ((ISettingsInstance)f.flowScope["edit"]).get<string>(p.id), (f, v) => ((ISettingsInstance)f.flowScope["edit"]).set(p.id, v),
                                                                   400f + QOLMod.VerticalScrollbarWidth(), GameFont.Small, QOLMod.LineHeight(GameFont.Small) * 3)
                            { EmptyIsNull = ((ISettingsFieldPropertiesPrimitiveText)p).EmptyIsNull });
                    }
                    else
                    {
                        rows.AddChild(new ColumnLayoutRenderer()
                            .AddChild(new LabelRenderer(p.label, 200f).attachTooltip(p))
                            .AddChild(new EditTextRenderer(f => ((ISettingsInstance)f.flowScope["edit"]).get<string>(p.id), (f, v) => ((ISettingsInstance)f.flowScope["edit"]).set(p.id, v),
                                                                   200f, ((ISettingsFieldPropertiesPrimitiveString)p).maxNumOfChars, ((ISettingsFieldPropertiesPrimitiveString)p).inputValidator)
                            { EmptyIsNull = ((ISettingsFieldPropertiesPrimitiveString)p).EmptyIsNull })
                            .AddChild(new ColumnLayoutRenderer().setPaddingBetweenColumns(0f)
                                .AddChild(new HideCollapseRenderer(f => ((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id), new ButtonRenderer(Materials.radioButtonOff, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsConfigured(p.id), 24f, 24f)))
                                .AddChild(new HideCollapseRenderer(f => !((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id) || !((ISettingsInstance)f.flowScope["edit"]).hasFieldConfiguredStatus(), new ButtonRenderer(Materials.radioButtonOn, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsNotConfigured(p.id), 24f, 24f)))
                            )
                           );
                    }
				} else if ( p.type == typeof(Color) ) {
					rows.AddChild(new ColumnLayoutRenderer()
				            .AddChild(new LabelRenderer(p.label, 200f).attachTooltip(p))
				            .AddChild(new EditColorRenderer(f=>((ISettingsInstance)f.flowScope["edit"]).get<Color>(p.id), 
				                                         (f,c) =>((ISettingsInstance)f.flowScope["edit"]).set<Color>(p.id,c),
				                                         200f))
				            .AddChild(new ColumnLayoutRenderer().setPaddingBetweenColumns(0f)
				            	.AddChild(new HideCollapseRenderer(f=>((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id),new ButtonRenderer(Materials.radioButtonOff, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsConfigured(p.id), 24f, 24f)))
			            		.AddChild(new HideCollapseRenderer(f=>!((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id)||!((ISettingsInstance)f.flowScope["edit"]).hasFieldConfiguredStatus(),new ButtonRenderer(Materials.radioButtonOn, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsNotConfigured(p.id), 24f, 24f)))
			            	)
				           );
				} else if ( p.isList && ((ISettingsFieldPropertiesList)p).isListSettings ) {
                    if (p is SettingsFieldPropertiesListSettings)
                    {
                        SettingsFieldPropertiesListSettings pList = (SettingsFieldPropertiesListSettings)p;
                        SettingsProperties props2 = SettingsStorage.getSettingsProperties(pList.idEnclosedSettings);
                        string deepEditPageId = currentPage + "edit" + pList.id + props2.id;
                        rows.AddChild(new ColumnLayoutRenderer()
                                .AddChild(new LabelRenderer(p.label, 400f, GameFont.Small, TextAnchor.MiddleCenter).attachTooltip(p))
                                .AddChild(new EmptyRenderer(0f))
                                .AddChild(new ColumnLayoutRenderer().setPaddingBetweenColumns(0f)
                                    .AddChild(new HideCollapseRenderer(f => ((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id), new ButtonRenderer(Materials.radioButtonOff, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsConfigured(p.id), 24f, 24f)))
                                    .AddChild(new HideCollapseRenderer(f => !((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id) || !((ISettingsInstance)f.flowScope["edit"]).hasFieldConfiguredStatus(), new ButtonRenderer(Materials.radioButtonOn, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsNotConfigured(p.id), 24f, 24f)))
                                )
                               );

                        bool sortable = pList.isSortable;

                        ListRenderer lstRenderer = new ListRenderer(f =>
                        {
                            List<SettingsInstance> lst = ((ISettingsInstance)f.flowScope["edit"]).get<List<SettingsInstance>>(pList.id);
                            //avoiding nullpointer if the default value for a list is null or if a save state didnt contain the list
                            if (lst == null)
                            {
                                lst = new List<SettingsInstance>();
                                ((ISettingsInstance)f.flowScope["edit"]).set<List<SettingsInstance>>(pList.id, lst);
                            }
                            return lst;
                        }, 400f + QOLMod.VerticalScrollbarWidth(), QOLMod.LineHeight(GameFont.Small) * 3)
                        { resizeVertically = QOLMod.getSettings().growLists };
                        rows.AddChild(lstRenderer);

                        lstRenderer.AddChild(new ButtonTextRenderer(f => ((ISettingsInstance)((IterationItem)f.pageScope["curItem"]).curItem).getLabel(),
                                                                       f =>
                                                                       {
                                                                           f.navigate(deepEditPageId);
                                                                           object curItem = ((IterationItem)f.pageScope["curItem"]).curItem;
                                                                           f.addPostRenderCallback(fl => fl.flowScope["edit"] = curItem);
                                                                       }, GameFont.Small, (pList.isRemovalAllowed ? 380f : 400f) - (sortable ? 40f : 0f)));

                        if ( sortable ) addSortingButtons(lstRenderer, p);

                        lstRenderer.AddChild(pList.isRemovalAllowed ? new ButtonTextRenderer("-", f =>
                                      {
                                          ISettingsInstance setting = (ISettingsInstance)f.flowScope["edit"];
                                          SettingsInstance lstItem = (SettingsInstance)((IterationItem)f.pageScope["curItem"]).curItem;
                                          setting.markFieldAsConfigured(p.id, true, false);
                                          f.addPostRenderCallback(fl =>
                                          {
                                              if (setting.get<List<SettingsInstance>>(p.id).Remove(lstItem)) setting.markAsChanged();
                                          });
                                      }, GameFont.Small, 20f) : null);

                        rows.AddChild(pList.isAddingAllowed ? new ColumnLayoutRenderer()
                                      .AddChild(new ButtonTextRenderer("+",
                                                                       f =>
                                                                       {
                                                                           SettingsInstance newItem = (SettingsInstance)pList.createNewEntry();
                                                                           if (newItem != null)
                                                                           {
                                                                               ((ISettingsInstance)f.flowScope["edit"]).get<List<SettingsInstance>>(p.id).Add(newItem);
                                                                               newItem.attach(((ISettingsInstance)f.flowScope["edit"]), pList);
                                                                               ((ISettingsInstance)f.flowScope["edit"]).markFieldAsConfigured(p.id, true, false);
                                                                               ((ISettingsInstance)f.flowScope["edit"]).markAsChanged();
                                                                           }
                                                                       },
                                                                       GameFont.Small, 400f)) : null);
                        pages.Add(deepEditPageId, editPage(props2, currentPage, deepEditPageId, pages, (f, nextPage) =>
                        {
                            if (nextPage.onBack() != deepEditPageId)
                            {
                                f.flowScope["edit"] = ((ISettingsInstance)f.flowScope["edit"]).getParentObject();
                            }
                        }));
                    }
                    else if (p is SettingsFieldPropertiesListSettingsReferences)
                    {
                        SettingsFieldPropertiesListSettingsReferences pList = (SettingsFieldPropertiesListSettingsReferences)p;
                        SettingsProperties props2 = SettingsStorage.getSettingsProperties(pList.idEnclosedSettings);
                        string deepEditPageId = currentPage + "edit" + pList.id + props2.id;
                        rows.AddChild(new ColumnLayoutRenderer()
                                .AddChild(new LabelRenderer(p.label, 400f, GameFont.Small, TextAnchor.MiddleCenter).attachTooltip(p))
                                .AddChild(new EmptyRenderer(0f))
                                .AddChild(new ColumnLayoutRenderer().setPaddingBetweenColumns(0f)
                                    .AddChild(new HideCollapseRenderer(f => ((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id), new ButtonRenderer(Materials.radioButtonOff, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsConfigured(p.id), 24f, 24f)))
                                    .AddChild(new HideCollapseRenderer(f => !((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id) || !((ISettingsInstance)f.flowScope["edit"]).hasFieldConfiguredStatus(), new ButtonRenderer(Materials.radioButtonOn, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsNotConfigured(p.id), 24f, 24f)))
                                )
                               );

                        bool sortable = pList.isSortable;

                        ListRenderer lstRenderer = new ListRenderer(f =>
                        {
                            List<SettingsInstance> lst = ((ISettingsInstance)f.flowScope["edit"]).get<List<SettingsInstance>>(pList.id);
                            //avoiding nullpointer if the default value for a list is null or if a save state didnt contain the list
                            if (lst == null)
                            {
                                lst = new List<SettingsInstance>();
                                ((ISettingsInstance)f.flowScope["edit"]).set<List<SettingsInstance>>(pList.id, lst);
                            }
                            return lst;
                        }, 400f + QOLMod.VerticalScrollbarWidth(), QOLMod.LineHeight(GameFont.Small) * 3)
                        { resizeVertically = QOLMod.getSettings().growLists };

                        rows.AddChild(lstRenderer);

                        lstRenderer.AddChild(new ButtonTextRenderer(f => ((ISettingsInstance)((IterationItem)f.pageScope["curItem"]).curItem).getLabel(),
                                                    f =>
                                                    {
                                                        f.navigate(deepEditPageId);
                                                        object curItem = ((IterationItem)f.pageScope["curItem"]).curItem;
                                                        f.addPostRenderCallback(fl => fl.flowScope["edit"] = curItem);
                                                    }, GameFont.Small, (pList.isRemovalAllowed ? 380f : 400f) - (sortable ? 40f : 0f)));

                        if ( sortable ) addSortingButtons(lstRenderer, p);

                        if (pList.isRemovalAllowed)
                        {
                            lstRenderer.AddChild(new ButtonTextRenderer("-", f =>
                            {
                                ISettingsInstance setting = (ISettingsInstance)f.flowScope["edit"];
                                SettingsInstance lstItem = (SettingsInstance)((IterationItem)f.pageScope["curItem"]).curItem;
                                setting.markFieldAsConfigured(p.id, true, false);
                                f.addPostRenderCallback(fl =>
                                {
                                    if (setting.get<List<SettingsInstance>>(p.id).Remove(lstItem)) setting.markAsChanged();
                                });
                            }, GameFont.Small, 20f));
                        }

                        rows.AddChild(pList.isAddingAllowed ? new ColumnLayoutRenderer()
                                      .AddChild(new ButtonTextRenderer("+",
                                                                       f =>
                                                                       {
                                                                           List<SettingsInstance> choices = SettingsStorage.getSettingsInstancesForSelection((ISettingsInstance)f.flowScope["edit"], pList);
                                                                           Find.WindowStack.Add(new Dialog_Select<SettingsInstance>(
                                                                               si=> {
                                                                                   ((ISettingsInstance)f.flowScope["edit"]).get<List<SettingsInstance>>(p.id).Add(si);
                                                                                   si.attach(((ISettingsInstance)f.flowScope["edit"]), pList);
                                                                                   ((ISettingsInstance)f.flowScope["edit"]).markFieldAsConfigured(p.id, true, false);
                                                                                   ((ISettingsInstance)f.flowScope["edit"]).markAsChanged();
                                                                                   return true;
                                                                               }, choices, si=>si.getLabel(), null, null));
                                                                       },
                                                                       GameFont.Small, 400f)) : null);
                        pages.Add(deepEditPageId, editPage(props2, currentPage, deepEditPageId, pages, (f, nextPage) =>
                        {
                            if (nextPage.onBack() != deepEditPageId)
                            {
                                f.flowScope["edit"] = ((ISettingsInstance)f.flowScope["edit"]).getParentObject();
                            }
                        }));
                    }
				} else if ( p.isList && ((ISettingsFieldPropertiesList)p).isListPrimitive ) {
					SettingsFieldPropertiesListPrimitive pList = (SettingsFieldPropertiesListPrimitive)p;
					Type listItemType = pList.type.GetGenericArguments()[0];
					rows.AddChild(new ColumnLayoutRenderer()
				            .AddChild(new LabelRenderer(p.label, 400f, GameFont.Small, TextAnchor.MiddleCenter).attachTooltip(p))
				            .AddChild(new EmptyRenderer(0f))
				            .AddChild(new ColumnLayoutRenderer().setPaddingBetweenColumns(0f)
				            	.AddChild(new HideCollapseRenderer(f=>((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id),new ButtonRenderer(Materials.radioButtonOff, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsConfigured(p.id), 24f, 24f)))
			            		.AddChild(new HideCollapseRenderer(f=>!((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id)||!((ISettingsInstance)f.flowScope["edit"]).hasFieldConfiguredStatus(),new ButtonRenderer(Materials.radioButtonOn, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsNotConfigured(p.id), 24f, 24f)))
			            	)
				           );
					ListRenderer lstRenderer = new ListRenderer(f => {
					                                                  	IList lst = ((ISettingsInstance)f.flowScope["edit"]).get<IList>(pList.id);
					                                                  	//avoiding nullpointer if the default value for a list is null or if a save state didnt contain the list
					                                                  	if ( lst == null ) {
                                                                            lst = (IList)Activator.CreateInstance(typeof(List<>).MakeGenericType(listItemType),null);
					                                                  		((ISettingsInstance)f.flowScope["edit"]).setUnchanged<IList>(pList.id,lst);
                                                                        }
					                                                  	return lst;
		                                                              },400f+QOLMod.VerticalScrollbarWidth(),QOLMod.LineHeight(GameFont.Small)*3) { resizeVertically = QOLMod.getSettings().growLists };

                    bool sortable = pList.isSortable;
                    float editRendererWidth;
					if ( pList.listLabelProducer != null ) {
						lstRenderer.AddChild(new LabelRenderer(f => pList.listLabelProducer((ISettingsInstance)f.flowScope["edit"],f), 200f));
						editRendererWidth = pList.isRemovalAllowed ? 180f : 200f;
					} else editRendererWidth = pList.isRemovalAllowed ? 380f : 400f;
                    if (sortable) editRendererWidth = editRendererWidth - 40f;
                    //Currently no min max etc for the list item type
                    if ( listItemType == typeof(float) ) {
						lstRenderer.AddChild(new EditNumericRenderer<float>(f=>(float)((IterationItem)f.pageScope["curItem"]).curItem,
			                                                      (f,v) => {
						                                          	ISettingsInstance si = ((ISettingsInstance)f.flowScope["edit"]);
						                                          	int indx = ((IterationItem)f.pageScope["curItem"]).index;
						                                          	f.addPostRenderCallback(fl => { si.get<IList>(pList.id)[indx] = v; si.markFieldAsConfigured(p.id, true, false); si.markAsChanged(); });
                                  	                        	  }, editRendererWidth/*,((ISettingsFieldPropertiesPrimitiveFloat)p).minValue,((ISettingsFieldPropertiesPrimitiveFloat)p).maxValue*/));
					} else if ( listItemType == typeof(int) ) {
						lstRenderer.AddChild(new EditNumericRenderer<int>(f=>(int)((IterationItem)f.pageScope["curItem"]).curItem,
				                                                  (f,v) => {
						                                          	ISettingsInstance si = ((ISettingsInstance)f.flowScope["edit"]);
						                                          	int indx = ((IterationItem)f.pageScope["curItem"]).index;
						                                          	f.addPostRenderCallback(fl => { si.get<IList>(pList.id)[indx] = v; si.markFieldAsConfigured(p.id, true, false); si.markAsChanged(); });
                                  	                        	  }, editRendererWidth/*,((ISettingsFieldPropertiesPrimitiveInt)p).minValue,((ISettingsFieldPropertiesPrimitiveInt)p).maxValue*/));
					} else if ( listItemType == typeof(bool) ) {
						lstRenderer.AddChild(new CheckboxRenderer(f=>(bool)((IterationItem)f.pageScope["curItem"]).curItem,
						                                          (f,v) => {
																	ISettingsInstance si = ((ISettingsInstance)f.flowScope["edit"]);
						                                          	int indx = ((IterationItem)f.pageScope["curItem"]).index;
						                                          	f.addPostRenderCallback(fl => { si.get<IList>(pList.id)[indx] = v; si.markFieldAsConfigured(p.id, true, false); si.markAsChanged(); });
						                                          }));
					} else if ( listItemType == typeof(string) ) {
						lstRenderer.AddChild(new EditTextRenderer(f=>(string)((IterationItem)f.pageScope["curItem"]).curItem,
						                                          (f,v) => {
						                                          	ISettingsInstance si = ((ISettingsInstance)f.flowScope["edit"]);
						                                          	int indx = ((IterationItem)f.pageScope["curItem"]).index;
						                                          	f.addPostRenderCallback(fl => { si.get<IList>(pList.id)[indx] = v; si.markFieldAsConfigured(p.id, true, false); si.markAsChanged(); });
                                  	                        	  }, editRendererWidth/*,((ISettingsFieldPropertiesPrimitiveString)p).maxNumOfChars,((ISettingsFieldPropertiesPrimitiveString)p).inputValidator*/));
					}


                    if (sortable) addSortingButtons(lstRenderer, p);

                    lstRenderer.AddChild(pList.isRemovalAllowed ? new ButtonTextRenderer("-", f => {
                     	ISettingsInstance setting = (ISettingsInstance)f.flowScope["edit"];
                     	int indx = setting.get<IList>(p.id).IndexOf(((IterationItem)f.pageScope["curItem"]).curItem);
                     	if ( indx >= 0 ) {
                     		setting.markFieldAsConfigured(p.id, true, false);
	                     	f.addPostRenderCallback(fl => {
								IList lst = setting.get<IList>(p.id);
								if (indx < lst.Count) {
									lst.RemoveAt(indx);
									setting.markAsChanged();
								}});
                     	}},
                     	GameFont.Small, 20f):null);
					
					rows.AddChild(lstRenderer);
					
					rows.AddChild(pList.isAddingAllowed ? new ColumnLayoutRenderer()
			              .AddChild(new ButtonTextRenderer("+", 
	                           f=> {
	                           	object newItem = pList.createNewEntry();
	                           	if ( newItem != null ) {
	                           		ISettingsInstance currentSettings = ((ISettingsInstance)f.flowScope["edit"]);
	                           		IList lst = currentSettings.get<IList>(p.id);
	                           		currentSettings.markFieldAsConfigured(p.id, true, false);
	                           		f.addPostRenderCallback(fl=>{
       		                        		lst.Add(newItem);
			                               	if ( pList.isSettings ) ((ISettingsInstance)newItem).attach(currentSettings, pList);
			                               	currentSettings.markAsChanged();
           		                        });
	                           	}
	                           },
	                           GameFont.Small, 400f)) : null);
				} else if ( p.isList && ((ISettingsFieldPropertiesList)p).isSelectable ) {
					SettingsFieldPropertiesListSelectable pList = (SettingsFieldPropertiesListSelectable)p;
					Type listItemType = pList.type.GetGenericArguments()[0];
					rows.AddChild(new ColumnLayoutRenderer()
				            .AddChild(new LabelRenderer(p.label, 400f, GameFont.Small, TextAnchor.MiddleCenter).attachTooltip(p))
				            .AddChild(new EmptyRenderer(0f))
				            .AddChild(new ColumnLayoutRenderer().setPaddingBetweenColumns(0f)
				            	.AddChild(new HideCollapseRenderer(f=>((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id),new ButtonRenderer(Materials.radioButtonOff, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsConfigured(p.id), 24f, 24f)))
			            		.AddChild(new HideCollapseRenderer(f=>!((ISettingsInstance)f.flowScope["edit"]).isFieldConfigured(p.id)||!((ISettingsInstance)f.flowScope["edit"]).hasFieldConfiguredStatus(),new ButtonRenderer(Materials.radioButtonOn, f => ((ISettingsInstance)f.flowScope["edit"]).markFieldAsNotConfigured(p.id), 24f, 24f)))
			            	)
				           );
                    ListRenderer lstRenderer = new ListRenderer(f => {
                        IList lst = ((ISettingsInstance)f.flowScope["edit"]).get<IList>(pList.id);
                        //avoiding nullpointer if the default value for a list is null or if a save state didnt contain the list
                        if (lst == null) {
                            lst = (IList)Activator.CreateInstance(typeof(List<>).MakeGenericType(listItemType), null);
                            ((ISettingsInstance)f.flowScope["edit"]).setUnchanged<IList>(pList.id, lst);
                        }
                        return lst;
                    }, 400f + QOLMod.VerticalScrollbarWidth(), QOLMod.LineHeight(GameFont.Small) * 3) { resizeVertically = QOLMod.getSettings().growLists };

                    bool sortable = pList.isSortable;

					lstRenderer.AddChild(new ButtonTextRenderer(f => pList.labelProducer(((IterationItem)f.pageScope["curItem"]).curItem),
					                                            f => Find.WindowStack.Add(new Dialog_Select<object>(
					                                            	wrapSelection(pList.onSelectItem(f),((ISettingsInstance)f.flowScope["edit"]),p.id), 
					                                            	pList.selectableItems(f), pList.labelProducer, null, pList.tooltipProducer)),
                                                       		GameFont.Small,pList.isRemovalAllowed ? 380f - (sortable ? 40f : 0f) : 400f - (sortable ? 40f : 0f)));
					if ( pList.tooltipProducer != null ) {
						lstRenderer.childs[0].attachTooltip(f => pList.tooltipProducer(((IterationItem)f.pageScope["curItem"]).curItem));
					}

                    if (sortable) addSortingButtons(lstRenderer, p);

					lstRenderer.AddChild(pList.isRemovalAllowed ? new ButtonTextRenderer("-", f => {
                     	ISettingsInstance setting = (ISettingsInstance)f.flowScope["edit"];
                     	setting.markFieldAsConfigured(p.id, true, false);
                     	object lstItem = ((IterationItem)f.pageScope["curItem"]).curItem;
                     	f.addPostRenderCallback(fl => {
							IList lst = setting.get<IList>(p.id);
							if (lst.Contains(lstItem)) {
								lst.Remove(lstItem);
								setting.markAsChanged();
							}});},
                     	GameFont.Small, 20f):null);
					
					rows.AddChild(lstRenderer);
					
					rows.AddChild(pList.isAddingAllowed ? new ColumnLayoutRenderer()
			              .AddChild(new ButtonTextRenderer("+", 
	                           f=> {
	                           	object newItem = pList.createNewEntry();
	                           	if ( newItem != null ) {
	                           		ISettingsInstance currentSettings = ((ISettingsInstance)f.flowScope["edit"]);
	                           		IList lst = currentSettings.get<IList>(p.id);
	                           		currentSettings.markFieldAsConfigured(p.id, true, false);
	                           		f.addPostRenderCallback(fl=>{
       		                        		lst.Add(newItem);
			                               	if ( pList.isSettings ) ((ISettingsInstance)newItem).attach(currentSettings, pList);
			                               	currentSettings.markAsChanged();
           		                        });
	                           	}
	                           },
	                           GameFont.Small, 400f)) : null);
				}
				
				Func<Flow,bool> isHidden = p.isHidden;
				if ( isHidden != null ) {
					int stopChildCount = rows.childs.Count;
					for ( int i = startChildCount; i < stopChildCount; i++ ) {
						HideCollapseRenderer hr = new HideCollapseRenderer(isHidden, rows.childs[i]);
						rows.childs[i] = hr;
					}
				}
			}
			
			return new PageRenderer(((ISettingsVisualizationEdit)props.visualization).getEditPageTitle, null, ()=>backPage, 
                (f, p) => { if ( f.GetPage(backPage) == p ) ((ISettingsInstance)f.flowScope["edit"]).OnStopEditing(); onNavigationHandler?.Invoke(f, p); }, 
                (f, p)=> { if (f.GetPage(backPage) == p) ((ISettingsInstance)f.flowScope["edit"]).OnStartEditing(); }).AddChild(new VerticalScrollViewRenderer().AddChild(rows));
		}

        protected void addSortingButtons(ListRenderer lstRenderer, ISettingsFieldProperties p)
        {
            lstRenderer.AddChild(new ButtonTextRenderer("A", f => {
                ISettingsInstance setting = (ISettingsInstance)f.flowScope["edit"];
                setting.markFieldAsConfigured(p.id, true, false);
                object lstItem = ((IterationItem)f.pageScope["curItem"]).curItem;
                f.addPostRenderCallback(fl => {
                    IList lst = setting.get<IList>(p.id);
                    int indx = lst.IndexOf(lstItem);
                    if (indx > 0)
                    {
                        lst[indx] = lst[indx - 1];
                        lst[indx - 1] = lstItem;
                        setting.markAsChanged();
                    }
                });
            }, GameFont.Small, 20f));
            lstRenderer.AddChild(new ButtonTextRenderer("V", f => {
                ISettingsInstance setting = (ISettingsInstance)f.flowScope["edit"];
                setting.markFieldAsConfigured(p.id, true, false);
                object lstItem = ((IterationItem)f.pageScope["curItem"]).curItem;
                f.addPostRenderCallback(fl => {
                    IList lst = setting.get<IList>(p.id);
                    int indx = lst.IndexOf(lstItem);
                    if (indx >= 0 && indx < lst.Count - 1)
                    {
                        lst[indx] = lst[indx + 1];
                        lst[indx + 1] = lstItem;
                        setting.markAsChanged();
                    }
                });
            }, GameFont.Small, 20f));
        }
		
		protected Func<object,bool> wrapSelection(Func<object,bool> callback, ISettingsInstance setting, string fieldId) {
			return obj => {
				bool res = callback(obj);
	        	if ( res ) {
	        		setting.markFieldAsConfigured(fieldId); 
	        	}
	        	return res;
			};
		}
		
		protected void addDefaultTraitCost(Dictionary<string, Dictionary<int, float>> dic, string defName, int degree, float cost) {
			Dictionary<int, float> dicDegrees;
			if ( !dic.TryGetValue(defName, out dicDegrees) ) {
				dicDegrees = new Dictionary<int, float>();
				dic.Add(defName,dicDegrees);
			}
			dicDegrees.Add(degree,cost);
		}
	}
}
