﻿/*
 * Created by SharpDevelop.
 * User: Malte Schulze
 * Date: 30.10.2017
 * Time: 06:21
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using UnityEngine;
using Verse;

namespace RIMMSqol.genericSettings
{
	/// <summary>
	/// Description of SettingsField.
	/// </summary>
	public interface ISettingsField
	{
		bool set(object value);
		object get();
		void OnBeforeSerialize(SettingsFieldProperties p);
		void OnAfterDeserialize(SettingsFieldProperties p);
		void ExposeData(SettingsFieldProperties p);
	}
	
	public class PrimitiveSettingsField : ISettingsField {
		protected object _value;
		public bool set(object value) {
			if ( (value == null && _value != null) || (value != null && !value.Equals(_value)) ) {
				_value = value;
				return true;
			}
			return false;
		}
		public object get() {
			return _value;
		}
		public void OnBeforeSerialize(SettingsFieldProperties p) {}
		public void OnAfterDeserialize(SettingsFieldProperties p) {}
		public void ExposeData(SettingsFieldProperties p) {
			MethodInfo mi = typeof(Scribe_Values).GetMethod("Look", BindingFlags.Public | BindingFlags.Static).MakeGenericMethod(p.type);
			object[] arguments = new object[]{_value,p.id,p.defaultValue,false};
			mi.Invoke(null,arguments);
			_value = arguments[0];
		}
	}
	
	public class SettingsSettingsField : ISettingsField {
		protected ISettingsInstance _value;
		public bool set(object value) {
			if ( (value == null && _value != null) || (value != null && !value.Equals(_value)) ) {
				_value = (ISettingsInstance)value;
				return true;
			}
			return false;
		}
		public object get() {
			return _value;
		}
		public void OnBeforeSerialize(SettingsFieldProperties p) {}
		public void OnAfterDeserialize(SettingsFieldProperties p) {}
		public void ExposeData(SettingsFieldProperties p) {
			MethodInfo miLook = null;
			string[] parameterNames = new string[]{"target","label","ctorArgs"};
			foreach ( MethodInfo mi in typeof(Scribe_Deep).GetMember("Look") ) {
				if ( mi.GetParameters().Length == parameterNames.Length ) {
					bool mismatch = false; int i = 0;
					foreach ( ParameterInfo pi in mi.GetParameters() ) {
						if ( !pi.Name.Equals(parameterNames[i]) ) {
							mismatch = true; break;
						} else {
							i++;
						}
					}
					if ( !mismatch ) {
						miLook = mi.MakeGenericMethod(p.type);
						break;
					}
				}
			}
			if ( miLook != null ) {
				object[] arguments = new object[]{_value,p.id,new object[0]};
				miLook.Invoke(null,arguments);
				_value = (ISettingsInstance)arguments[0];
			} else {
				Log.Error("Failed to find generic Scribe_Deep.Look for saving settings.");
			}
		}
	}
	
	public class DefSettingsField : ISettingsField {
		public static Type GetTypeFromAnyAssemblyVersion(string typeName)
		{
			try {
				Type t = Type.GetType(typeName);
				if ( t == null ) {
					AssemblyName an = new AssemblyName(typeName.Substring(typeName.IndexOf(',')+1));
					an.Version = null;
					Assembly ass = Assembly.Load(an);
					if ( ass != null ) {
						if ( !ass.ReflectionOnly )
							return ass.GetType(typeName.Substring(0,typeName.IndexOf(',')));
						else Log.Warning("Assembly for type "+typeName+" was reflection only and cannot be used.");
					}
					return null;
				}
		        return t;
			} catch (Exception ex) {
				Log.Error("Error while trying to load type " + typeName + ".\n"+ex.Message+"\n"+ex.StackTrace);
				return null;
			}
		}
		
		protected string _defName, _typeName;
		protected Def _def;
		public bool set(object value)
		{
			if ((value != null && _defName != ((Def)value).defName) || (value == null && _defName != null)) { 
				_def = (Def)value;
				_typeName = value != null ? value.GetType().AssemblyQualifiedName : null;
				_defName = value != null ? ((Def)value).defName : null; 
				return true; 
			}
			return false;
		}
		public object get()
		{
			if ( _def == null && _defName != null ) {
				MethodInfo mi = typeof(DefDatabase<>).MakeGenericType(GetTypeFromAnyAssemblyVersion(_typeName)).GetMethod("GetNamed", BindingFlags.Public | BindingFlags.Static);
				_def = (Def)mi.Invoke(null,new object[]{_defName,false});
			}
			return _def;
		}
		public void OnBeforeSerialize(SettingsFieldProperties p) {}
		public void OnAfterDeserialize(SettingsFieldProperties p) {}
		public void ExposeData(SettingsFieldProperties p) {
			Scribe_Values.Look<string>(ref _defName, p.id, null);
			Scribe_Values.Look<string>(ref _typeName, p.id+"_type", null);
		}
	}

    public class IconSettingsField : ISettingsField
    {
        protected KeyValuePair<string, Texture2D>? value;
        public bool set(object value)
        {
            if (value == null)
            {
                this.value = null;
                return true;
            }
            if (value is KeyValuePair<string, Texture2D>?)
            {
                this.value = (KeyValuePair<string, Texture2D>?)value;
                return true;
            }
            if (value is KeyValuePair<string, Texture2D>)
            {
                this.value = (KeyValuePair<string, Texture2D>)value;
                return true;
            }
            string iconPath = value as string;
            if (!this.value.HasValue || this.value == null || this.value.Value.Key != iconPath)
            {
                this.value = new KeyValuePair<string, Texture2D>(iconPath, null);
                return true;
            }
            return false;
        }
        public object get()
        {
            if (this.value.HasValue)
            {
                string s = this.value.Value.Key;
                if ( s != null && this.value.Value.Value == null && UnityData.IsInMainThread )
                {
                    this.value = new KeyValuePair<string, Texture2D>(s, ContentFinder<Texture2D>.Get(s, false));
                    if (this.value.Value.Value == null) this.value = null;
                }
            }
            return this.value;
        }
        public void OnBeforeSerialize(SettingsFieldProperties p) { }
        public void OnAfterDeserialize(SettingsFieldProperties p) { }
        public void ExposeData(SettingsFieldProperties p)
        {
            string iconPath;
            if (this.value.HasValue) iconPath = this.value.Value.Key;
            else iconPath = null;
            Scribe_Values.Look<string>(ref iconPath, p.id, (string)p.defaultValue);
            if ( Scribe.mode == LoadSaveMode.LoadingVars )
            {
                set(iconPath);
            }
        }
    }

    public class TypeSettingsField : ISettingsField {		
		protected string _typeName;
		protected Type _type;
		public bool set(object value)
		{
			if ((value != null && _typeName != ((Type)value).AssemblyQualifiedName) || (value == null && _typeName != null)) { 
				_type = (Type)value;
				_typeName = value != null ? ((Type)value).AssemblyQualifiedName : null; 
				return true; 
			}
			return false;
		}
		public object get()
		{
			if ( _type == null && _typeName != null ) {
				_type = DefSettingsField.GetTypeFromAnyAssemblyVersion(_typeName);
			}
			return _type;
		}
		public void OnBeforeSerialize(SettingsFieldProperties p) {}
		public void OnAfterDeserialize(SettingsFieldProperties p) {}
		public void ExposeData(SettingsFieldProperties p) {
			Scribe_Values.Look<string>(ref _typeName, p.id, null);
		}
	}
	
	public class ListSettingsField : ISettingsField {
		protected IList _unresolvedValues;
		protected Type _innerType;
		protected IList _value;
		public bool set(object value)
		{
			IList localValue = (IList)get();
			/*if ( typeof(Def).IsAssignableFrom(_innerType) ) {
				Log.Message("Def list is being set");
				Log.Message("old");
				if ( localValue != null ) {
					foreach ( object o in localValue ) {
						Log.Message(""+o);
					}
				} else Log.Message("IsNull");
				Log.Message("new");
				if ( value != null ) {
					foreach ( object o in (IList)value ) {
						Log.Message(""+o);
					}
				} else Log.Message("IsNull");
			}*/
			if ( !listEquals(localValue,(IList)value) ) {
				if ( localValue != null ) {
					localValue.Clear();
				} else if (value != null) {
					localValue = (IList)Activator.CreateInstance(value.GetType());
				}
				if ( value != null ) {
					foreach ( object o in (IList)value ) {
						localValue.Add(o);
					}
					_value = localValue;
				} else {
					_value = null;
					_unresolvedValues = null;
				}
				return true;
			}
			return false;
		}
		public object get()
		{
			/*if ( typeof(Def).IsAssignableFrom(_innerType) ) {
				Log.Message("Def list");
				Log.Message("Value");
				if ( _value != null ) {
					foreach ( object o in _value ) {
						Log.Message(""+o);
					}
				} else Log.Message("IsNull");
				Log.Message("Unresolved Value");
				if ( _unresolvedValues != null ) {
					foreach ( object o in _unresolvedValues ) {
						Log.Message(""+o);
					}
				} else Log.Message("IsNull");
			}*/
			if ( _value == null && _unresolvedValues != null ) {
				if ( typeof(Def).IsAssignableFrom(_innerType) ) {
					//Log.Message("transferring unresolved to value");
					MethodInfo mi = typeof(DefDatabase<>).MakeGenericType(_innerType).GetMethod("GetNamed", BindingFlags.Public | BindingFlags.Static);
					IList resolvedValue = (IList)Activator.CreateInstance(typeof(List<>).MakeGenericType(_innerType));
					foreach ( object o in _unresolvedValues ) {
                        object resolved;
                        if (o != null)
                        {
                            resolved = mi.Invoke(null, new object[] { o, false });
                            if (resolved == null)
                            {
                                Log.Error("Failed to load referenced def " + o + " in RIMMSqol setting. Maybe a missing mod?");
                            }
                        }
                        else resolved = null;
                        resolvedValue.Add(resolved);
						//Log.Message("Adding def to value list: "+o+" ("+resolved+")");
					}
					_value = resolvedValue;
				} else _value = _unresolvedValues;
			}
			return _value;
		}
		protected bool listEquals(IList current, IList value) {
			if ( current == value ) return true;
			if ( current == null || value == null ) return false;
			if ( current.Count != value.Count ) return false;
			IEnumerator enumerator1 = current.GetEnumerator();
			IEnumerator enumerator2 = value.GetEnumerator();
			while ( enumerator1.MoveNext() ) {
				if ( !enumerator2.MoveNext() ) return false;
				if ( object.ReferenceEquals(enumerator1.Current, enumerator2.Current) ) continue;
				if ( object.ReferenceEquals(null,enumerator1.Current) || object.ReferenceEquals(null,enumerator2.Current) ) return false;
				if ( !enumerator1.Current.Equals(enumerator2.Current) ) return false;
			}
			return true;
		}
		public void OnBeforeSerialize(SettingsFieldProperties p){}
		public void OnAfterDeserialize(SettingsFieldProperties p){}
		public void ExposeData(SettingsFieldProperties p)
		{
			_innerType = p.type.GetGenericArguments()[0];
			if (Scribe.mode == LoadSaveMode.Saving) {
				if ( _value != null ) {
					//Log.Message("saving list and transferring changes to unresolved before saving unresolved.");
					//if value contains a change it won't be null 
					if ( typeof(Def).IsAssignableFrom(_innerType) ) {
						_unresolvedValues = (IList)Activator.CreateInstance(typeof(List<>).MakeGenericType(typeof(string)));
						foreach ( object o in _value ) {
                            if ( o != null ) _unresolvedValues.Add(((Def)o).defName);
                            else _unresolvedValues.Add(null);
                        }
					} else _unresolvedValues = _value;
				}
			}
			
			MethodInfo miLook = null;
			string[] parameterNames = new string[]{"list","label","lookMode","ctorArgs"};
			foreach ( MethodInfo mi in typeof(Scribe_Collections).GetMember("Look") ) {
				if ( mi.GetParameters().Length == parameterNames.Length ) {
					bool mismatch = false; int i = 0;
					foreach ( ParameterInfo pi in mi.GetParameters() ) {
						if ( !pi.Name.Equals(parameterNames[i]) ) {
							mismatch = true; break;
						} else {
							i++;
						}
					}
					if ( !mismatch ) {
						miLook = mi;
						break;
					}
				}
			}
			if ( miLook != null ) {
				if ( typeof(Def).IsAssignableFrom(_innerType) )
					miLook = miLook.MakeGenericMethod(typeof(string));
				else miLook = miLook.MakeGenericMethod(_innerType);
				object[] arguments = new object[]{_unresolvedValues,p.id,LookMode.Undefined,null};
				miLook.Invoke(null,arguments);
				_unresolvedValues = (IList)arguments[0];
				_value = null;
				/*Log.Message("unresolved values inner type: "+_innerType);
				Log.Message("unresolved values is def list: "+typeof(Def).IsAssignableFrom(_innerType));
				foreach (object o in _unresolvedValues) {
					Log.Message(""+o);
				}*/
			} else {
				Log.Error("Failed to build generic look method for list field!");
			}
		}
	}
}
