﻿/*
 * Created by SharpDevelop.
 * User: Malte Schulze
 * Date: 09.02.2019
 * Time: 15:41
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;
using System.Linq;
using RimWorld;
using RimWorld.Planet;
using UnityEngine;
using Verse;

namespace RIMMSqol.remnantcolony
{
	/// <summary>
	/// Description of QOLCityDialog_ExchangeCitizens.
	/// </summary>
	public class QOLCityDialog_ExchangeCitizens : Window {
		protected QOLCity city;
		protected Pawn negotiator;
		protected Map map;
		protected Vector2 scrollPositionLeft, scrollPositionRight;
		protected List<Pawn> selectedPawns = new List<Pawn>(), colonyPawns = new List<Pawn>(), cityPawns = new List<Pawn>();
		protected const float GapToBorder = 0f, GapBetweenSelectionBoxes = 16f, GapBetweenPortraitAndLabel = 4f, GapBetweenSelectionAndButtons = 8f;
		protected readonly static Vector2 portraitSize = new Vector2(70f,70f), GapBetweenPortraits = new Vector2(4f,4f);
		
		override public Vector2 InitialSize {
			get {
				return new Vector2(640f, 300f);
			}
		}
		
		public QOLCityDialog_ExchangeCitizens(QOLCity city, Pawn negotiator) {
			this.closeOnCancel = true;
			this.doCloseX = true;
			this.onlyOneOfTypeAllowed = true;
			this.absorbInputAroundWindow = true;
			this.forcePause = true;
			this.city = city;
			this.negotiator = negotiator;
			if ( this.negotiator != null ) this.map = this.negotiator.Map;
			else this.map = null;
		}
		
		public override void PreOpen() {
			base.PreOpen();
			selectedPawns.Clear();
			scrollPositionLeft = Vector2.zero;
			scrollPositionRight = Vector2.zero;
			colonyPawns.Clear();
			cityPawns.Clear();
			cityPawns.AddRange(city.population.citizens.Keys);
			ReloadMap();
		}
		
		public void DoSelectionBox(Rect outRect, ref Vector2 scrollPosition, List<Pawn> pawns) {
			float LabelHeight = QOLMod.LineHeight(GameFont.Tiny);
			float rowHeight = LabelHeight+portraitSize.y+GapBetweenPortraitAndLabel+GapBetweenPortraits.y;
			float clmnWidth = portraitSize.x+20f+GapBetweenPortraits.x;
			Vector2 offsetPortrait = new Vector2((clmnWidth-portraitSize.x)/2f,0f);
			
			int columns = Math.Max(1,Mathf.FloorToInt(outRect.width / clmnWidth));
			Rect viewRect = new Rect(0,0,columns * clmnWidth,Mathf.Max(outRect.height,((pawns.Count/columns)+(pawns.Count%columns!=0?1:0))*rowHeight));
			viewRect.width -= QOLMod.VerticalScrollbarWidth();
			
			Widgets.BeginScrollView(outRect, ref scrollPosition, viewRect);
			GameFont oldFont = Text.Font;
			TextAnchor oldAnchor = Text.Anchor;
			try {
				Text.Font = GameFont.Tiny;
				Text.Anchor = TextAnchor.MiddleCenter;
				Rect portraitRect;
				Rect labelRect;
				List<Pawn>.Enumerator enumerator = pawns.GetEnumerator();
				bool finished = false;
				for ( int row = 0; !finished; row++ ) {
					for ( int clmn = 0; clmn < columns; clmn++ ) {
						if ( !enumerator.MoveNext() ) {
							finished = true;
							break;
						}
						Pawn p = enumerator.Current;
						portraitRect = new Rect(clmn*clmnWidth+offsetPortrait.x,row*rowHeight+offsetPortrait.y,portraitSize.x,portraitSize.y);
						labelRect = new Rect(clmn*clmnWidth,row*rowHeight+rowHeight-LabelHeight-GapBetweenPortraits.y,clmnWidth-GapBetweenPortraits.x,LabelHeight);
						
						if (Mouse.IsOver(portraitRect)||Mouse.IsOver(labelRect)||IsSelected(p)) {
							Widgets.DrawHighlight(portraitRect);
							Widgets.DrawHighlight(labelRect);
						}
						
						if (Mouse.IsOver(portraitRect)||Mouse.IsOver(labelRect)) {
							if (Event.current.type == EventType.MouseDown && Event.current.button == 0) {
								Event.current.Use();
								GUIUtility.keyboardControl = 0;
								Select(p);
							}
						}
						
						RenderTexture tx = PortraitsCache.Get(p, portraitSize, Rot4.South);
						GUI.DrawTexture(portraitRect, tx);
						Widgets.Label(labelRect, p.LabelShort);
					}
				}
			} finally {
				Widgets.EndScrollView();
				Text.Font = oldFont;
				Text.Anchor = oldAnchor;
			}
		}
		
		protected bool IsSelected(Pawn p) {
			return selectedPawns.Contains(p);
		}
		
		protected void Select(Pawn p) {
			if ( !selectedPawns.Remove(p) ) selectedPawns.Add(p);
		}
		
		protected void ReloadMap() {
			colonyPawns.Clear();
			IEnumerable<Pawn> availablePawns;
			if ( map != null ) {
				availablePawns = map.mapPawns.AllPawnsSpawned;
			} else if ( negotiator != null ) {
				Caravan caravan = negotiator.GetCaravan();
				if ( caravan != null ) {
					availablePawns = caravan.PawnsListForReading;
				} else {
					availablePawns = Enumerable.Empty<Pawn>();
				}
			} else {
				availablePawns = Enumerable.Empty<Pawn>();
			}
			foreach ( Pawn p in availablePawns ) {
				if ( ((p.Faction != null && p.Faction.IsPlayer)||(p.IsPrisoner && p.guest.HostFaction != null && p.guest.HostFaction.IsPlayer)) && 
				    city.defCity.population.IsImmigrationAllowed(city,p) ) {
					colonyPawns.Add(p);
				}
			}
			selectedPawns.RemoveAll(p=>!cityPawns.Contains(p));
			colonyPawns.Remove(negotiator);
			scrollPositionLeft = Vector2.zero;
		}
		
		protected void TransferSelectedPawns() {
			List<Pawn> moveToCity = new List<Pawn>(), moveToColony = new List<Pawn>();
			foreach ( Pawn p in colonyPawns ) {
				if ( IsSelected(p) ) {
					if ( city.AddCitizen(p) ) {
						//Passing to world and setting faction should be done by the city?
						if (p.Spawned) {
							p.DeSpawn(DestroyMode.Vanish);
						}
						if (p.ownership != null) {
							p.ownership.UnclaimAll();
						}
						/*if (p.guest != null) {
							p.guest.SetGuestStatus(null, false);
						}*/
						if ( negotiator != null ) {
							Caravan caravan = negotiator.GetCaravan();
							if ( caravan != null ) caravan.RemovePawn(p);
						}
						p.inventory.UnloadEverything = false;
						if ( !p.IsWorldPawn() ) {
							Faction fac = p.Faction;
							Find.WorldPawns.PassToWorld(p, PawnDiscardDecideMode.Decide);
							p.SetFaction(fac);
						}
						
						moveToCity.Add(p);
					}
				}
			}
			foreach ( Pawn p in cityPawns ) {
				if ( IsSelected(p) ) {
					if ( map != null ) TradeUtility.SpawnDropPod(DropCellFinder.TradeDropSpot(map), map, p);
					else if ( negotiator != null ) {
						Caravan caravan = negotiator.GetCaravan();
						if ( caravan != null ) caravan.AddPawn(p, true);
					}
					city.RemoveCitizen(p);
					
					moveToColony.Add(p);
				}
			}
			colonyPawns.RemoveAll(moveToCity.Contains);
			cityPawns.RemoveAll(moveToColony.Contains);
			//when using drop pods we have to wait until the pawn actually spawns in. so for the menu a direct reversal is not possible.
			//colonyPawns.AddRange(moveToColony);
			cityPawns.AddRange(moveToCity);
			city.Refresh();
			Reset();
		}
		
		protected void Reset() {
			selectedPawns.Clear();
			scrollPositionLeft = Vector2.zero;
			scrollPositionRight = Vector2.zero;
		}
		
		public override void DoWindowContents(Rect inRect) {
			float ButtonHeight = QOLMod.LineHeight(GameFont.Medium) + 4f;
			GameFont oldFont = Text.Font;
			TextAnchor oldAnchor = Text.Anchor;
			inRect = inRect.ContractedBy(GapToBorder);
			
			Rect leftSelectionBox = new Rect(inRect.x,inRect.y,inRect.width/2f-GapBetweenSelectionBoxes/2f,inRect.height-ButtonHeight-GapBetweenSelectionAndButtons);
			Rect rightSelectionBox = new Rect(leftSelectionBox.xMax+GapBetweenSelectionBoxes,leftSelectionBox.y,leftSelectionBox.width,leftSelectionBox.height);
			DoSelectionBox(leftSelectionBox, ref scrollPositionLeft, colonyPawns);
			DoSelectionBox(rightSelectionBox, ref scrollPositionRight, cityPawns);
			
			Text.Font = GameFont.Medium;
			Text.Anchor = TextAnchor.MiddleCenter;
			Rect btnRect = new Rect(inRect.x,leftSelectionBox.yMax+GapBetweenSelectionAndButtons,inRect.width/3f,ButtonHeight);
			if ( Widgets.ButtonText(btnRect,"Reset") ) {
				Reset();
			}
			btnRect.x += btnRect.width;
			if ( Widgets.ButtonText(btnRect,"Transfer") ) {
				TransferSelectedPawns();
				Close(true);
			}
			btnRect.x += btnRect.width;
			if ( Widgets.ButtonText(btnRect,"Close") ) {
				Close(true);
			}
			btnRect.x += btnRect.width;
			
			Text.Font = oldFont;
			Text.Anchor = oldAnchor;
		}
		
		public override void PostClose() {
			base.PostClose();
			selectedPawns.Clear();
		}
	}
}
