﻿/*
 * Created by SharpDevelop.
 * User: Malte Schulze
 * Date: 02.11.2019
 * Time: 05:21
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;
using System.Linq;
using RIMMSqol.genericSettings;
using RIMMSqol.renderers;
using RimWorld;
using RimWorld.Planet;
using Verse;

namespace RIMMSqol.storyteller
{
	/// <summary>
	/// Description of StorytellerTreeNode_FilterTarget.
	/// </summary>
	public class StorytellerTreeNode_FilterTarget : StorytellerTreeNode {
		public List<string> targetTypes = new List<string>();
		public List<IncidentTargetTagDef> targetTags;
		public bool whitelist = true;
		public bool filterBiomesOnDiseaseChance = false;
		public string pick = "all";
		public StoryDanger minDanger = StoryDanger.None;
		public SimpleCurve chanceByThreatPointsCurve;
		
		protected Func<IIncidentTarget, bool> filter;
		protected Dictionary<IIncidentTarget,bool> skipTargetsForThisInterval = new Dictionary<IIncidentTarget,bool>();
		protected int lastInterval;
		protected List<float> curvePoints;
		
		public StorytellerTreeNode_FilterTarget() : this(null) { }
		
		public StorytellerTreeNode_FilterTarget(StorytellerTreeNode parent) : base(parent) {
			// disable once ConvertClosureToMethodGroup
			filter = t => allowsTarget(t);
		}
		
		protected virtual bool IsOfTargetType(IIncidentTarget target, string targetType, Map m, Caravan c) {
			switch (targetType) {
				case "TempMap": if ( m != null && m.IsTempIncidentMap ) return true; break;
				case "Map[player]": if ( m != null && m.IsPlayerHome ) return true; break;
				case "Map[friendly]": if ( m != null && m.ParentFaction != null && !m.ParentFaction.HostileTo(Faction.OfPlayer) ) return true; break;
				case "Map[hostile]": if ( m != null && m.ParentFaction != null && m.ParentFaction.HostileTo(Faction.OfPlayer) ) return true; break;
				case "Caravan[player]": if ( c != null && c.IsPlayerControlled ) return true; break;
				case "Caravan[friendly]": if ( c != null && c.Faction != null && !c.Faction.HostileTo(Faction.OfPlayer) ) return true; break;
				case "Caravan[hostile]": if ( c != null && c.Faction != null && c.Faction.HostileTo(Faction.OfPlayer) ) return true; break;
				case "World": if ( target is World ) return true; break;
			}
			return false;
		}
		
		protected bool allowsTarget(IIncidentTarget target) {
			if ( target == null ) return false;
			
			bool skipTarget;
			if ( skipTargetsForThisInterval.TryGetValue(target, out skipTarget) ) {
				return skipTarget;
			}
			
			Map m = target as Map;
			Caravan c = target as Caravan;
			
			if ( filterBiomesOnDiseaseChance && target.Tile != -1 ) {
				BiomeDef biome = Find.WorldGrid[target.Tile].biome;
				float mtb = biome.diseaseMtbDays;
				mtb *= Find.Storyteller.difficulty.diseaseIntervalFactor;
				if ( !Rand.MTBEventOccurs(mtb, 60000f, 1000f) ) {
					skipTargetsForThisInterval[target] = false;
					return false;
				}
			}
			
			if ( minDanger != StoryDanger.None ) {
				if (m == null || m.dangerWatcher.DangerRating < minDanger) {
					skipTargetsForThisInterval[target] = false;
					return false;
				}
			}
			
			if ( chanceByThreatPointsCurve != null && chanceByThreatPointsCurve.PointsCount > 0 ) {
				if ( Rand.Value > chanceByThreatPointsCurve.Evaluate(StorytellerUtility.DefaultThreatPointsNow(target)) ) {
					skipTargetsForThisInterval[target] = false;
					return false;
				}
			}
			
			if ( !targetTags.NullOrEmpty() ) {
				if ( target.IncidentTargetTags().Intersect(targetTags).Any() ) {
					skipTargetsForThisInterval[target] = whitelist;
					return whitelist;
				}
			}
			
			if ( !targetTypes.NullOrEmpty() ) {
				foreach ( string allowedTargetType in targetTypes ) {
					if ( IsOfTargetType(target, allowedTargetType, m, c) ) {
						skipTargetsForThisInterval[target] = whitelist;
						return whitelist;
					}
				}
			}
			
			skipTargetsForThisInterval[target] = !whitelist;
			return !whitelist;
		}
		
		protected int measureSize(IIncidentTarget t) {
			Map m = t as Map;
			if ( m != null ) {
				return m.mapPawns.ColonistsSpawnedCount;
			}
			Caravan c = t as Caravan;
			if ( c != null ) {
				return c.PlayerPawnsForStoryteller.Count();
			}
			return 0;
		}
		
		public override IEnumerable<StorytellerTreeIncident> TryStartIncident(StorytellerTree tree) {
			if ( lastInterval != tree.lastInterval ) {
				lastInterval = tree.lastInterval;
				skipTargetsForThisInterval.Clear();
			}
			StorytellerTreeComp_TargetFilter targetFilterComp = tree.getComp<StorytellerTreeComp_TargetFilter>();
			targetFilterComp.ApplyFilter(filter);
			
			Func<IIncidentTarget, bool> filterPick;
			if ( pick != "all" ) {
				IIncidentTarget pickedTarget = null;
				switch ( pick ) {
					case "random": pickedTarget = targetFilterComp.FilteredTargets.RandomElementWithFallback(null); break;
					case "largest": pickedTarget = targetFilterComp.FilteredTargets.MaxByWithFallback(measureSize, null); break;
					case "first": pickedTarget = targetFilterComp.FilteredTargets.FirstOrDefault(); break;
					case "last": pickedTarget = targetFilterComp.FilteredTargets.LastOrDefault(); break;
				}
				if ( pickedTarget != null ) {
					filterPick = t => t == pickedTarget;
				} else filterPick = t => false;
				targetFilterComp.ApplyFilter(filterPick);
			} else filterPick = null;
			
			foreach ( StorytellerTreeIncident fi in base.TryStartIncident(tree) ) yield return fi;
			
			if ( filterPick != null ) targetFilterComp.RemoveFilter(filterPick);
			targetFilterComp.RemoveFilter(filter);
			
			yield break;
		}
		
		public override void ExposeData() {
			base.ExposeData();
			Scribe_Collections.Look<string>(ref targetTypes, "spacingIntervalsMin");
			Scribe_Collections.Look<IncidentTargetTagDef>(ref targetTags, "targetTags");
			Scribe_Values.Look<bool>(ref whitelist, "whitelist", true);
			Scribe_Values.Look<string>(ref pick, "pick", "all");
			Scribe_Values.Look<bool>(ref filterBiomesOnDiseaseChance, "filterBiomesOnDiseaseChance", false);
			Scribe_Values.Look<StoryDanger>(ref minDanger, "minDanger", StoryDanger.None);
			
			if (Scribe.mode == LoadSaveMode.Saving) {
				curvePoints = curvePointsAsList();
			}
			Scribe_Collections.Look<float>(ref curvePoints, "chanceByThreatPointsCurve");
			if (Scribe.mode == LoadSaveMode.ResolvingCrossRefs) {
				loadCurveFromPoints(curvePoints, out chanceByThreatPointsCurve);
				if ( curvePoints != null ) curvePoints.Clear(); 
				curvePoints = null;
			}
		}
		
		public void loadCurvePoints(List<float> curvePoints) {
			loadCurveFromPoints(curvePoints, out chanceByThreatPointsCurve);
		}
		
		public List<float> curvePointsAsList() {
			return curvePointsAsList(chanceByThreatPointsCurve);
		}
		
		protected override void InitSettings(SettingsProperties p) {
			base.InitSettings(p);
			p.fields.Add(new SettingsFieldPropertiesPrimitiveBool<StorytellerTreeNode_FilterTarget>("Whitelist","Whitelist",10,true,t=>t.whitelist));
			p.fields.Add(new SettingsFieldPropertiesPrimitiveBool<StorytellerTreeNode_FilterTarget>("FilterBiomesOnDiseaseChance","Filter Biomes on Disease Chance",15,true,t=>t.filterBiomesOnDiseaseChance));
			p.fields.Add(new SettingsFieldPropertiesSelectable("Pick","Pick Target",20,typeof(string),"all",t=>((StorytellerTreeNode_FilterTarget)t).pick,
			                                                   f=>new []{"all","random","largest","first","last"},
			                                                   f=>entry=>{ ((ISettingsInstance)f.flowScope["edit"]).set<string>("Pick",(string)entry); return true;},
			                                                   o=>(string)o,false));
			p.fields.Add(new SettingsFieldPropertiesListSelectable("TargetTypes","Target Types",30,typeof(List<string>),new List<string>(),
				h=>((StorytellerTreeNode_FilterTarget)h).targetTypes == null ? new List<string>() : ((StorytellerTreeNode_FilterTarget)h).targetTypes.ToList(),
				f=>new []{"TempMap","Map[player]","Map[friendly]","Map[hostile]","Caravan[player]","Caravan[friendly]","Caravan[hostile]","World"},
				flow=>{
					ISettingsInstance setting = (ISettingsInstance)flow.flowScope["edit"];
					int listIndex = ((IterationItem)flow.pageScope["curItem"]).index;
					return listEntry=>{
					flow.addPostRenderCallback(fl=>{setting.get<List<string>>("TargetTypes")[listIndex] = (string)listEntry; setting.markAsChanged();});
					return true;
					};},o=>(string)o, () => "Map[player]",true));
			p.fields.Add(new SettingsFieldPropertiesListSelectable("TargetTags","Target Tags",40,typeof(List<IncidentTargetTagDef>),new List<IncidentTargetTagDef>(),
				h=>((StorytellerTreeNode_FilterTarget)h).targetTags == null ? new List<IncidentTargetTagDef>() : ((StorytellerTreeNode_FilterTarget)h).targetTags.ToList(),
				f=>DefDatabase<IncidentTargetTagDef>.AllDefsListForReading.Cast<object>(),
				flow=>{
					ISettingsInstance setting = (ISettingsInstance)flow.flowScope["edit"];
					int listIndex = ((IterationItem)flow.pageScope["curItem"]).index;
					return listEntry=>{
					flow.addPostRenderCallback(fl=>{setting.get<List<IncidentTargetTagDef>>("TargetTags")[listIndex] = (IncidentTargetTagDef)listEntry; setting.markAsChanged();});
					return true;
					};},SettingsFieldPropertiesSelectable.DefLabelProducer, () => DefDatabase<IncidentTargetTagDef>.AllDefsListForReading.FirstOrDefault(),true));
			p.fields.Add(new SettingsFieldPropertiesSelectable("MinDanger","Min Danger",50,typeof(string),Enum.GetName(typeof(StoryDanger),StoryDanger.None),
			                                                   t=>Enum.GetName(typeof(StoryDanger),((StorytellerTreeNode_FilterTarget)t).minDanger),
                                                       f=>from StoryDanger d in Enum.GetValues(typeof(StoryDanger)) orderby Enum.GetName(typeof(StoryDanger),d) select (object)Enum.GetName(typeof(StoryDanger),d),
			                                             f=>def=>{ ((ISettingsInstance)f.flowScope["edit"]).set<string>("MinDanger",(string)def); return true; },
			                                             o=>(string)o,false));
			p.fields.Add(new SettingsFieldPropertiesListPrimitive("ChanceByThreatPointsCurve","Chance by Threat Curve",60,typeof(List<float>),null,
			                                                      t=>((StorytellerTreeNode_FilterTarget)t).curvePointsAsList(),null,()=>1.0f,true));
		}
		
		protected override void AdoptSettings(SettingsInstance si, StorytellerTreeNode obj) {
			base.AdoptSettings(si, obj);
			StorytellerTreeNode_FilterTarget sttn = (StorytellerTreeNode_FilterTarget)obj;
			sttn.whitelist = si.get<bool>("Whitelist");
         	sttn.filterBiomesOnDiseaseChance = si.get<bool>("FilterBiomesOnDiseaseChance");
         	sttn.pick = si.get<string>("Pick");
         	sttn.minDanger = (StoryDanger)Enum.Parse(typeof(StoryDanger),si.get<string>("MinDanger"));
         	List<string> targetTypes = si.get<List<string>>("TargetTypes");
        	if ( targetTypes != null ) {
        		if ( sttn.targetTypes == null ) sttn.targetTypes = new List<string>();
        		else sttn.targetTypes.Clear();
        		sttn.targetTypes.AddRange(targetTypes);
        	}
         	List<IncidentTargetTagDef> targetTags = si.get<List<IncidentTargetTagDef>>("TargetTags");
         	if ( targetTags != null ) {
         		if ( sttn.targetTags == null ) sttn.targetTags = new List<IncidentTargetTagDef>();
        		else sttn.targetTags.Clear();
        		sttn.targetTags.AddRange(targetTags);
         	}
         	sttn.loadCurvePoints(si.get<List<float>>("ChanceByThreatPointsCurve"));
		}
		
		public override void DoSettingsDialog(StorytellerTree tree) {
			Find.WindowStack.Add(new Dialog_EditObject<StorytellerTreeNode_FilterTarget>(this,GetSettings().id,AdoptSettings));
		}
	}
	
	public class StorytellerTreeNodeConfiguration_FilterTarget : StorytellerTreeNodeConfiguration {
		public List<string> targetTypes;
		public List<IncidentTargetTagDef> targetTags;
		public SimpleCurve chanceByThreatPointsCurve;
		public bool whitelist = true;
		public string pick = "all";
		public bool filterBiomesOnDiseaseChance = false;
		public StoryDanger minDanger = StoryDanger.None;
		
		public override StorytellerTreeNode CreateNode(StorytellerTreeNode parent) {
			StorytellerTreeNode_FilterTarget n = new StorytellerTreeNode_FilterTarget(parent);
			n.label = label;
			n.targetTypes = targetTypes != null ? new List<string>(targetTypes) : null;
			n.targetTags = targetTags != null ? new List<IncidentTargetTagDef>(targetTags) : null;
			n.whitelist = whitelist;
			n.pick = pick;
			n.filterBiomesOnDiseaseChance = filterBiomesOnDiseaseChance;
			if ( chanceByThreatPointsCurve != null && chanceByThreatPointsCurve.PointsCount > 0 ) {
				n.chanceByThreatPointsCurve = new SimpleCurve(); n.chanceByThreatPointsCurve.SetPoints(chanceByThreatPointsCurve.Points);
			} else n.chanceByThreatPointsCurve = null;
			n.minDanger = minDanger;
			return n;
		}
	}
}
